@@ -1,9 +1,9 @@
-#ident "$Name: release2_0-12 $"
-#ident "$Id: intercept.c,v 1.33 2002/03/15 00:59:51 ttsai Exp $"
+#ident "$Name: release2_0-16 $"
+#ident "$Id: intercept.c,v 1.40 2002/05/31 17:37:34 ttsai Exp $"
 
 
 /*
- * Copyright (C) 2000 Avaya Labs, Avaya Inc.
+ * Copyright (C) 2002 Avaya Labs, Avaya Inc.
  * Copyright (C) 1999 Bell Labs, Lucent Technologies.
  * Copyright (C) Arash Baratloo, Timothy Tsai, Navjot Singh, and Hamilton Slye.
  *
@@ -63,12 +63,16 @@
  */
 typedef void *(*memcpy_t) (void *dest, const void *src, size_t n);
 typedef char *(*strcpy_t) (char *dest, const char *src);
+typedef char *(*strncpy_t) (char *dest, const char *src, size_t n);
 typedef wchar_t *(*wcscpy_t) (wchar_t *dest, const wchar_t *src);
 typedef char *(*stpcpy_t) (char *dest, const char *src);
 typedef wchar_t *(*wcpcpy_t) (wchar_t *dest, const wchar_t *src);
 typedef char *(*strcat_t) (char *dest, const char *src);
+typedef char *(*strncat_t) (char *dest, const char *src, size_t n);
 typedef wchar_t *(*wcscat_t) (wchar_t *dest, const wchar_t *src);
 typedef int (*vsprintf_t) (char *str, const char *format, va_list ap);
+typedef int (*vsnprintf_t) (char *str, size_t size, const char *format, va_list
+	ap);
 typedef int (*vprintf_t) (const char *format, va_list ap);
 typedef int (*vfprintf_t) (FILE *fp, const char *format, va_list ap);
 typedef char *(*getwd_t) (char *buf);
@@ -176,6 +180,30 @@ char *strcpy(char *dest, const char *src)
     return dest;
 }
 
+char *strncpy(char *dest, const char *src, size_t n)
+{
+    static strncpy_t real_strncpy = NULL;
+    size_t max_size, len;
+
+    if (!real_strncpy)
+	real_strncpy = (strncpy_t) getLibraryFunction("strncpy");
+
+    if (_libsafe_exclude)
+	return real_strncpy(dest, src, n);
+
+    if ((max_size = _libsafe_stackVariableP(dest)) == 0) {
+	LOG(5, "strncpy(<heap var> , <src>)\n");
+	return real_strncpy(dest, src, n);
+    }
+
+    LOG(4, "strncpy(<stack var> , <src>) stack limit=%d)\n", max_size);
+
+    if (n > max_size && (len = strnlen(src, max_size)) == max_size)
+	_libsafe_die("Overflow caused by strncpy()");
+
+    return real_strncpy(dest, src, n);
+}
+
 char *stpcpy(char *dest, const char *src)
 {
     static stpcpy_t real_stpcpy = NULL;
@@ -208,6 +236,7 @@ char *stpcpy(char *dest, const char *src)
     return dest + len;
 }
 
+#ifndef MISSING_WCSNLEN
 wchar_t *wcscpy(wchar_t *dest, const wchar_t *src)
 {
     static wcscpy_t real_wcscpy = NULL;
@@ -289,6 +318,7 @@ wchar_t *wcpcpy(wchar_t *dest, const wchar_t *src)
      */
     return real_wcpcpy(dest, src);
 }
+#endif /* MISSING_WCSNLEN */
 
 /*
  * This is needed!  See the strcpy() for the reason. -ab.
@@ -347,6 +377,35 @@ char *strcat(char *dest, const char *src)
 }
 
 
+char *strncat(char *dest, const char *src, size_t n)
+{
+    static strncat_t real_strncat = NULL;
+    size_t max_size;
+    uint dest_len, src_len;
+
+    if (!real_strncat)
+	real_strncat = (strncat_t) getLibraryFunction("strncat");
+
+    if (_libsafe_exclude)
+	return real_strncat(dest, src, n);
+
+    if ((max_size = _libsafe_stackVariableP(dest)) == 0) {
+	LOG(5, "strncat(<heap var> , <src>)\n");
+	return real_strncat(dest, src, n);
+    }
+
+    LOG(4, "strncat(<stack var> , <src>) stack limit=%d\n", max_size);
+    dest_len = strnlen(dest, max_size);
+    src_len = strnlen(src, max_size);
+
+    if (dest_len + n > max_size && dest_len + src_len >= max_size)
+	_libsafe_die("Overflow caused by strncat()");
+
+    return real_strncat(dest, src, n);
+}
+
+
+#ifndef MISSING_WCSNLEN
 wchar_t *wcscat(wchar_t *dest, const wchar_t *src)
 {
     static wcscat_t real_wcscat = NULL;
@@ -377,6 +436,7 @@ wchar_t *wcscat(wchar_t *dest, const wchar_t *src)
 
     return dest;
 }
+#endif /* MISSING_WCSNLEN */
 
 
 /*
@@ -1130,12 +1190,15 @@ int _IO_vfprintf(FILE *fp, const char *format, va_list ap)
 int sprintf(char *str, const char *format, ...)
 {
     static vsprintf_t real_vsprintf = NULL;
+    static vsnprintf_t real_vsnprintf = NULL;
     size_t max_size;
     va_list ap;
     int res;
 
     if (!real_vsprintf)
 	real_vsprintf = (vsprintf_t) getLibraryFunction("vsprintf");
+    if (!real_vsnprintf)
+	real_vsnprintf = (vsnprintf_t) getLibraryFunction("vsnprintf");
 
     if (_libsafe_exclude) {
 	va_start(ap, format);
@@ -1163,7 +1226,8 @@ int sprintf(char *str, const char *format, ...)
      * max_size-1 is to make sure that there is room for the string-terminating
      * NULL.
      */
-    if ((res = vsnprintf(str, max_size, format, ap)) == -1 || res > max_size-1)
+    res = real_vsnprintf(str, max_size, format, ap);
+    if (res == -1 || res > max_size-1)
     {
 	_libsafe_die("overflow caused by sprintf()");
     }
@@ -1173,14 +1237,64 @@ int sprintf(char *str, const char *format, ...)
 }
 
 
+int snprintf(char *str, size_t size, const char *format, ...)
+{
+    static vsnprintf_t real_vsnprintf = NULL;
+    size_t max_size;
+    va_list ap;
+    int res;
+
+    if (!real_vsnprintf)
+	real_vsnprintf = (vsnprintf_t) getLibraryFunction("vsnprintf");
+
+    if (_libsafe_exclude) {
+	va_start(ap, format);
+	res = real_vsnprintf(str, size, format, ap);
+	va_end(ap);
+	return res;
+    }
+
+    if ((max_size = _libsafe_stackVariableP(str)) == 0) {
+	LOG(5, "snprintf(<heap var>, <format>)\n");
+	va_start(ap, format);
+	res = real_vsnprintf(str, size, format, ap);
+	va_end(ap);
+	return res;
+    }
+
+    LOG(4, "snprintf(<stack var>, <format>) stack limit=%d\n", max_size);
+    va_start(ap, format);
+
+    /*
+     * Some man pages say that -1 is returned if vsnprintf truncates the
+     * output.  However, some implementations actually return the number of
+     * chars that would have been output with a sufficiently large output
+     * buffer.  Hence, we check for both res==-1 and res>max_size-1.  The
+     * max_size-1 is to make sure that there is room for the string-terminating
+     * NULL.
+     */
+    res = real_vsnprintf(str, size, format, ap);
+    if ((res == -1 || res > max_size-1) && (size > max_size))
+    {
+	_libsafe_die("overflow caused by snprintf()");
+    }
+    va_end(ap);
+
+    return res;
+}
+
+
 int vsprintf(char *str, const char *format, va_list ap)
 {
     static vsprintf_t real_vsprintf = NULL;
+    static vsnprintf_t real_vsnprintf = NULL;
     size_t max_size;
     int res;
 
     if (!real_vsprintf)
 	real_vsprintf = (vsprintf_t) getLibraryFunction("vsprintf");
+    if (!real_vsnprintf)
+	real_vsnprintf = (vsnprintf_t) getLibraryFunction("vsnprintf");
 
     if (_libsafe_exclude)
 	return real_vsprintf(str, format, ap);
@@ -1200,7 +1314,8 @@ int vsprintf(char *str, const char *format, va_list ap)
      * max_size-1 is to make sure that there is room for the string-terminating
      * NULL.
      */
-    if ((res = vsnprintf(str, max_size, format, ap)) == -1 || res > max_size-1)
+    res = real_vsnprintf(str, max_size, format, ap);
+    if (res == -1 || res > max_size-1)
     {
 	_libsafe_die("overflow caused by vsprintf()");
     }
@@ -1208,6 +1323,42 @@ int vsprintf(char *str, const char *format, va_list ap)
 }
 
 
+int vsnprintf(char *str, size_t size, const char *format, va_list ap)
+{
+    static vsnprintf_t real_vsnprintf = NULL;
+    size_t max_size;
+    int res;
+
+    if (!real_vsnprintf)
+	real_vsnprintf = (vsnprintf_t) getLibraryFunction("vsnprintf");
+
+    if (_libsafe_exclude)
+	return real_vsnprintf(str, size, format, ap);
+
+    if ((max_size = _libsafe_stackVariableP(str)) == 0) {
+	LOG(5, "vsnprintf(<heap var>, <format>, <va_list>)\n");
+	return real_vsnprintf(str, size, format, ap);
+    }
+
+    LOG(4, "vsnprintf(<stack var>, <format>, <va_list>)\n");
+
+    /*
+     * Some man pages say that -1 is returned if vsnprintf truncates the
+     * output.  However, some implementations actually return the number of
+     * chars that would have been output with a sufficiently large output
+     * buffer.  Hence, we check for both res==-1 and res>max_size-1.  The
+     * max_size-1 is to make sure that there is room for the string-terminating
+     * NULL.
+     */
+    res = real_vsnprintf(str, size, format, ap);
+    if ((res == -1 || res > max_size-1) && (size > max_size))
+    {
+	_libsafe_die("overflow caused by vsnprintf()");
+    }
+    return res;
+}
+
+
 char *getwd(char *buf)
 {
     static getwd_t real_getwd = NULL;
