@@ -1,8 +1,8 @@
-#ident "$Name: release2_0-12 $"
-#ident "$Id: util.c,v 1.50 2002/03/14 14:14:11 ttsai Exp $"
+#ident "$Name: release2_0-16 $"
+#ident "$Id: util.c,v 1.58 2002/05/30 14:44:38 ttsai Exp $"
 
 /*
- * Copyright (C) 2000 Avaya Labs, Avaya Inc.
+ * Copyright (C) 2002 Avaya Labs, Avaya Inc.
  * Copyright (C) 1999 Bell Labs, Lucent Technologies.
  * Copyright (C) Arash Baratloo, Timothy Tsai, and Navjot Singh.
  *
@@ -751,6 +751,22 @@ void _libsafe_dump_stack(char *file, int linenum) {
 	LOG(1, "Dumping stack to %s.\n", filename);
     }
 
+
+
+
+/* For debugging only!!! */
+{
+    char cmd[1000];
+    sprintf(cmd, "cat /proc/%d/maps >/tmp/maps.%d", getpid(), getpid());
+    system(cmd);
+    printf("Copied maps to /tmp/maps.%d\n", getpid());
+}
+
+
+
+
+
+
     fprintf(fp, "Stack dump:  sp=%p  fp=%p  stack_start=%p\n",
 	sp, __builtin_frame_address(0), current_stack_start);
     fprintf(fp, "Initiating dump from file=%s linenum=%d\n", file, linenum);
@@ -964,6 +980,7 @@ static void sendmail(char *recipient, char *message) {
 	"Subject: ***** libsafe violation detected *****\r\n"
 	"To: %s\r\n"
 	"Date: %s\r\n"
+	"\r\n"
 	"Libsafe violation detected on %s at %s\r\n"
 	"%s\r\n",
 	recipient, ctime_nonewline(&t), hostname, ctime(&t), message))
@@ -1008,9 +1025,10 @@ static void get_stack_trace(const char *filename, char *buf, size_t size)
         pid = getpid();
         
         ret = snprintf(buf, size,
-                       "Detected an attempt to write across stack boundary.\n"
-                       "Terminating %s.\n"
-                       "    uid=%d  euid=%d  pid=%d\n", filename, getuid(), geteuid(), pid);
+		"Detected an attempt to write across stack boundary.\n"
+		"Terminating %s.\n"
+		"    uid=%d  euid=%d  pid=%d\n",
+		filename, getuid(), geteuid(), pid);
 
         /*
          * Print out the call stack.  We can assume that the stack is a normal
@@ -1020,7 +1038,8 @@ static void get_stack_trace(const char *filename, char *buf, size_t size)
         ret += snprintf(buf + ret, size - ret, "Call stack:\n");
         for ( fp = __builtin_frame_address(0); *fp; fp = *(void **) fp ) {
                 ra = *((caddr_t*)(fp+4));
-                ret += snprintf(buf + ret, size - ret, "    %p\n", (caddr_t)((uint)ra-5));
+                ret += snprintf(buf + ret, size - ret, "    %p\n",
+			(caddr_t)((uint)ra-5));
         }
 }
 
@@ -1100,6 +1119,8 @@ static int set_user_infos(idmef_user_t *user)
 
 static void fill_assessment(idmef_assessment_t *assessment) 
 {
+        idmef_action_t *action;
+        
         idmef_assessment_impact_new(assessment);
         
         assessment->impact->severity = impact_high;
@@ -1111,9 +1132,19 @@ static void fill_assessment(idmef_assessment_t *assessment)
         else
                 assessment->impact->type = user;
         
-        idmef_string_set_constant(&assessment->impact->description, "Stack overflow attempt detected and avoided by libsafe");
+        idmef_string_set_constant(&assessment->impact->description,
+		"Stack overflow attempt detected and avoided by libsafe");
         idmef_assessment_confidence_new(assessment);
         assessment->confidence->rating = high;
+
+        action = idmef_assessment_action_new(assessment);
+        if ( ! action )
+                return;
+
+        action->category = taken_offline;
+        idmef_string_set_constant(&action->description,
+		"Libsafe killed the target process in order to prevent the "
+		"overflow attack from succeeding");
 }
 
 
@@ -1167,7 +1198,8 @@ static void prelude_alert(char *filename)
         fill_assessment(alert->assessment);
         
         idmef_string_set_constant(&alert->analyzer.model, "Libsafe");
-        idmef_string_set_constant(&alert->analyzer.class, "Stack Overflow Detection Library");
+        idmef_string_set_constant(&alert->analyzer.class,
+		"Stack Overflow Detection Library");
         idmef_string_set_constant(&alert->analyzer.version, VERSION);
 
         /*
@@ -1221,7 +1253,8 @@ static void prelude_alert(char *filename)
         if ( ! classification )
                 goto err;
 
-        idmef_string_set_constant(&classification->name, "Stack Overflow Attempt");
+        idmef_string_set_constant(&classification->name,
+		"Stack Overflow Attempt");
 
         /*
          * Include the call trace.
@@ -1232,7 +1265,7 @@ static void prelude_alert(char *filename)
         
         data->type = string;
         idmef_string_set_constant(&data->meaning, "Stack overflow data");
-        idmef_string_set(&data->data, buf);
+        idmef_additional_data_set_data(data, string, buf, strlen(buf) + 1);
 
         /*
          * send the message synchronously (use 1 for asynchronous send).
@@ -1255,6 +1288,132 @@ static void prelude_alert(char *filename)
 #endif /* NOTIFY_WITH_PRELUDE */
 
 
+/*
+ * Sanity check for stack frame pointers.  Return 0 if the check passes.  Else
+ * return 1.
+ */
+static int check_nextfp(caddr_t fp, caddr_t nextfp) {
+    caddr_t stack_start = find_stack_start((void*)&fp);
+
+    /*
+     * The following checks are meant to detect code that doesn't insert
+     * frame pointers onto the stack.  (i.e., code that is compiled with
+     * -fomit-frame-pointer).
+     */
+
+    if (nextfp > stack_start) {
+	LOG(2, "fp > stack_start; bypass enabled\n");
+	_libsafe_exclude = 1;
+	return 1;
+    }
+
+    /*
+     * Make sure frame pointers are word aligned.
+     */
+    if ((uint)nextfp & 0x03) {
+	LOG(2, "fp not word aligned; bypass enabled\n");
+	_libsafe_exclude = 1;
+	return 1;
+    }
+
+    /*
+     * Make sure frame pointers are monotonically increasing.
+     */
+    if (nextfp <= fp) {
+	LOG(2, "fp not monotonically increasing; bypass enabled\n");
+	_libsafe_exclude = 1;
+	return 1;
+    }
+
+    return 0;
+}
+
+
+struct maps_st {
+    caddr_t start, end;
+    char    *path;
+};
+
+
+/*
+ * Read /proc/<pid>/maps to find the memory-mapped regions for this process.
+ */
+static int get_memory_maps(struct maps_st **mapsptr) {
+    struct maps_st  *maps = NULL;
+    char    filename[200], buf[500];
+    FILE    *fp;
+    int	    count, i;
+    char    *p;
+
+    snprintf(filename, sizeof(filename), "/proc/%d/maps", getpid());
+
+    /*
+     * First pass:  Find out how many memory regions there are.
+     */
+    if ((fp=fopen(filename, "r")) == NULL) {
+	return 0;
+    }
+
+    count = 0;
+    while (fgets(buf, sizeof(buf), fp)) {
+	count++;
+    }
+
+    fclose(fp);
+
+    maps = (struct maps_st *) malloc(count * sizeof(struct maps_st));
+    *mapsptr = maps;
+
+    /*
+     * Second pass:  Fill in the table with the region info.
+     */
+    if ((fp=fopen(filename, "r")) == NULL) {
+	if (maps) free(maps);
+	return 0;
+    }
+
+    i = 0;
+    while (fgets(buf, sizeof(buf), fp) && i < count) {
+	sscanf(buf, "%p-%p", &maps[i].start, &maps[i].end);
+	p = strchr(buf, '/');
+	if (p) {
+	    maps[i].path = strdup(p);
+	    
+	    /*
+	     * Strip off the trailing newline.
+	     */
+	    p = strchr(maps[i].path, '\n');
+	    if (p) *p = (char)NULL;
+	}
+	else
+	{
+	    maps[i].path = NULL;
+	}
+	i++;
+    }
+
+    fclose(fp);
+
+    return count;
+}
+
+
+/*
+ * Return the index into the maps array that corresponds to the caller_addr.
+ * If caller_addr is not in any of the regions in maps, then return -1.
+ */
+static int find_caller_addr(struct maps_st *maps, int count, caddr_t
+	caller_addr)
+{
+    int	i;
+
+    for (i=0; i<count; i++) {
+	if (caller_addr >= maps[i].start && caller_addr <= maps[i].end)
+	    return i;
+    }
+
+    return -1;
+}
 
 
 /*
@@ -1266,6 +1425,8 @@ void _libsafe_warn(char *format, ...)
 {
     char    exename[MAXPATHLEN];
     va_list args;
+    struct maps_st	*maps;
+    int		count, index;
 
     dying = 1;
 
@@ -1276,6 +1437,8 @@ void _libsafe_warn(char *format, ...)
     
     va_start(args, format);
 
+    count = get_memory_maps(&maps);
+
     /*
      * add an entry to syslog()
      */
@@ -1283,44 +1446,52 @@ void _libsafe_warn(char *format, ...)
     LOG(1, "Turned off syslog entries for debugging.\n");
 #else
     openlog(LIBNAME, LOG_CONS | LOG_PID, LOG_AUTHPRIV);
-    syslog(LOG_CRIT, "%s", VERSION);
+    syslog(LOG_CRIT, "Libsafe version %s", VERSION);
     syslog(LOG_CRIT, "Detected an attempt to write across stack boundary.");
     syslog(LOG_CRIT, "Terminating %s.", exename);
     syslog(LOG_CRIT, "    uid=%d  euid=%d  pid=%d", getuid(), geteuid(),
 	    getpid());
-    {
-	/*
-	 * Print out the call stack.  We can assume that the stack is a normal
-	 * stack, since _libsafe_stackVariableP(), _libsafe_raVariableP(), or
-	 * _libsafe_span_stack_frames() had to be called first.
-	 */
-	caddr_t	fp, ra;
-	syslog(LOG_CRIT, "Call stack:\n");
-	for (fp=__builtin_frame_address(0); *fp; fp=*(void **)fp) {
-	    ra = *((caddr_t*)(fp+4));
-	    syslog(LOG_CRIT, "    %p\n", (caddr_t)((uint)ra-5));
-	}
-    }
-    syslog(LOG_CRIT, format, args);
-//    closelog();
 #endif
-
+    LOG(1, "Libsafe version %s\n", VERSION);
     LOG(1, "Detected an attempt to write across stack boundary.\n");
     LOG(1, "Terminating %s.\n", exename);
     LOG(1, "    uid=%d  euid=%d  pid=%d\n", getuid(), geteuid(), getpid());
+
     {
 	/*
 	 * Print out the call stack.  We can assume that the stack is a normal
 	 * stack, since _libsafe_stackVariableP(), _libsafe_raVariableP(), or
 	 * _libsafe_span_stack_frames() had to be called first.
 	 */
-	caddr_t	fp, ra;
+	caddr_t	fp, ra, nextfp, caller_addr;
+#ifndef DEBUG_TURN_OFF_SYSLOG
+	syslog(LOG_CRIT, "Call stack:\n");
+#endif
 	LOG(1, "Call stack:\n");
-	for (fp=__builtin_frame_address(0); *fp; fp=*(void **)fp) {
+
+	for (fp=__builtin_frame_address(0); *fp; fp=nextfp) {
 	    ra = *((caddr_t*)(fp+4));
-	    LOG(1, "    %p\n", (caddr_t)((uint)ra-5));
+
+	    /*
+	     * Find the memory region and corresponding mapped file associated
+	     * with this address.
+	     */
+	    caller_addr = (caddr_t)((uint)ra-5);
+	    index = find_caller_addr(maps, count, caller_addr);
+
+#ifndef DEBUG_TURN_OFF_SYSLOG
+	    syslog(LOG_CRIT, "    %p  %s\n", caller_addr, maps[index].path);
+#endif
+	    LOG(1, "    %p\t%s\n", caller_addr, maps[index].path);
+
+	    nextfp = *(void **)fp;
+	    if (check_nextfp(fp, nextfp))
+		break;
 	}
     }
+#ifndef DEBUG_TURN_OFF_SYSLOG
+    syslog(LOG_CRIT, format, args);
+#endif
     if (1 <= LOG_LEVEL) {
 	vfprintf(stderr, format, args);
 	fprintf(stderr, "\n");
@@ -1352,7 +1523,7 @@ void _libsafe_warn(char *format, ...)
 	 * Form the descriptive message.
 	 */
 	snprintf(errmsg, sizeof(errmsg),
-	    "Libsafe %s\r\n"
+	    "Libsafe version %s\r\n"
 	    "Detected an attempt to write across stack boundary.\r\n"
 	    "Terminating %s.\r\n"
 	    "    uid=%d  euid=%d  pid=%d\r\n",
@@ -1366,15 +1537,33 @@ void _libsafe_warn(char *format, ...)
 	     * _libsafe_raVariableP(), or _libsafe_span_stack_frames() had to
 	     * be called first.
 	     */
-	    caddr_t	fp, ra;
+	    caddr_t	fp, ra, nextfp, caller_addr;
+	    struct maps_st	*maps;
+	    int		count, index;
+
+	    count = get_memory_maps(&maps);
+
 	    snprintf(buf, sizeof(buf), "Call stack:\r\n");
 	    strncat(errmsg, buf,
 		sizeof(errmsg) - strnlen(errmsg,sizeof(errmsg)) - 1);
-	    for (fp=__builtin_frame_address(0); *fp; fp=*(void **)fp) {
+	    for (fp=__builtin_frame_address(0); *fp; fp=nextfp) {
 		ra = *((caddr_t*)(fp+4));
-		snprintf(buf, sizeof(buf), "    %p\r\n", (caddr_t)((uint)ra-5));
+
+		/*
+		 * Find the memory region and corresponding mapped file
+		 * associated with this address.
+		 */
+		caller_addr = (caddr_t)((uint)ra-5);
+		index = find_caller_addr(maps, count, caller_addr);
+
+		snprintf(buf, sizeof(buf), "    %p\t%s\r\n", caller_addr,
+			maps[index].path);
 		strncat(errmsg, buf,
 		    sizeof(errmsg) - strnlen(errmsg,sizeof(errmsg)) - 1);
+
+		nextfp = *(void **)fp;
+		if (check_nextfp(fp, nextfp))
+		    break;
 	    }
 	}
 #ifdef DUMP_STACK
@@ -1428,6 +1617,9 @@ void _libsafe_warn(char *format, ...)
     }
 #endif	/* NOTIFY_WITH_EMAIL */
 
+    if (maps)
+	free(maps);
+
     /*
      * Since we are justing doing a warning, set dying=0 to indicate that we
      * should resume libsafe checking now.
